from typing import Dict, Optional

from pykechain.enums import (
    FontAwesomeMode,
    SidebarAccessLevelOptions,
    SidebarItemAlignment,
    SidebarType,
    URITarget,
)
from pykechain.exceptions import IllegalArgumentError
from pykechain.models.sidebar.sidebar_base import SideBarItem


class SideBarButton(SideBarItem):
    """
    Side-bar button class.

    Every custom button in the side-bar is maintained as an object of this class.
    The original KE-chain buttons for the project detail, tasks and work breakdown
    structure are not separate buttons.

    :cvar allowed_attributes: allowed additional attributed provided as options alongside
        the specifically allowed ones.
    :cvar item_type: the item type of this class. Defaults to a BUTTON.
    """

    _allowed_attributes = [
        "displayName_nl",
        "displayName_en",
        "displayName_de",
        "displayName_fr",
        "displayName_it",
    ]
    _item_type = SidebarType.BUTTON

    def __init__(
        self,
        side_bar_manager: "SideBarManager",
        json: Optional[Dict] = None,
        title: Optional[str] = None,
        icon: Optional[str] = None,
        uri: Optional[str] = None,
        alignment: SidebarItemAlignment = SidebarItemAlignment.TOP,
        minimum_access_level: SidebarAccessLevelOptions = SidebarAccessLevelOptions.IS_MEMBER,
        uri_target: URITarget = URITarget.INTERNAL,
        icon_mode: FontAwesomeMode = FontAwesomeMode.REGULAR,
        **kwargs,
    ):
        """
        Create a side-bar button.

        :param side_bar_manager: Manager object to which the button is linked.
        :param json: the json response to construct the :class:`SideBarButton` from
        :param title: visible label of the button
        :param icon: FontAwesome icon of the button
        :param uri: Uniform Resource Identifier, the address of the linked page
        :param uri_target: type of URI, either internal or external
        :param alignment: alignment of the button top or bottom
        :param minimum_access_level: the minimum permission needed to see the button
        :param icon_mode: FontAwesome display mode of the icon
        :returns None
        :raises IllegalArgumentError: When the provided Argument is not according to the type.
        """
        super().__init__()

        if json is None:
            json = {}

        title = title if title else json.get("displayName")
        icon = icon if icon else json.get("displayIcon")
        uri = uri if uri else json.get("uri")
        uri_target = json.get("uriTarget", uri_target)
        icon_mode = json.get("displayIconMode", icon_mode)
        alignment = json.get("align", alignment)
        minimum_access_level = json.get("minimumAccessLevel", minimum_access_level)

        if not isinstance(title, str):
            raise IllegalArgumentError(f'title must be a string, "{title}" is not.')
        if not isinstance(icon, str):
            raise IllegalArgumentError(f'icon must be a string, "{icon}" is not.')
        if not isinstance(uri, str):
            raise IllegalArgumentError(f'uri must be a string, "{uri}" is not.')
        if uri_target not in URITarget.values():
            raise IllegalArgumentError(
                f'uri_target must be a URITarget option, "{uri_target}" is not.'
            )
        if icon_mode not in FontAwesomeMode.values():
            raise IllegalArgumentError(
                f'icon_mode must be a FontAwesomeMode option, "{icon_mode}" is not.'
            )

        for key in kwargs.keys():
            if key not in self._allowed_attributes:
                raise IllegalArgumentError(
                    f'Attribute "{key}" is not supported in the configuration of a side-bar'
                    " card."
                )

        self._manager: "SideBarManager" = side_bar_manager
        self.display_name: str = title
        self.display_icon: str = icon
        self.uri: str = uri
        self.uri_target: URITarget = uri_target
        self.display_icon_mode: FontAwesomeMode = icon_mode
        self.alignment: SidebarItemAlignment = alignment
        self.minimum_access_level: SidebarAccessLevelOptions = minimum_access_level

        self._other_attributes = kwargs
        for key in self._allowed_attributes:
            if key in json:
                self._other_attributes[key] = json[key]

    def as_dict(self) -> Dict:
        """
        Retrieve the configuration data, or `meta`, of the side-bar button.

        :return: dictionary of the configuration data
        :rtype dict
        """
        config = {
            "itemType": self._item_type,
            "displayName": self.display_name,
            "displayIcon": self.display_icon,
            "uriTarget": self.uri_target,
            "uri": self.uri,
            "displayIconMode": self.display_icon_mode,
            "align": self.alignment,
            "minimumAccessLevel": self.minimum_access_level,
        }
        config.update(self._other_attributes)
        config = {k: v for k, v in config.items() if v is not None}

        return config



"""A python library to connect and interact with KE-chain."""
import sys

from .__about__ import version
from .client import Client
from .helpers import get_project

__all__ = ("Client", "get_project", "version")

if sys.version_info.major == 2 or (
    sys.version_info.major == 3 and sys.version_info.minor < 7
):
    raise RuntimeError(
        "Python version >= `3.7` is required for this version of `pykechain` to operate. "
        "Please use `pykechain` version `3.x.x` for usage in combination with Python `3.6`"
    )



import copy

from pykechain.enums import PropertyType, PropertyVTypes, _AllRepresentations

#
# Validators and Validator Effects
#
from pykechain.utils import UUID_REGEX_PATTERN

effects_jsonschema_stub = {
    "type": "object",
    "additionalProperties": False,
    "required": ["effect", "config"],
    "properties": {
        "effect": {"type": ["string", "null"]},
        "config": {"type": "object"},
    },
}

validator_jsonschema_stub = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "type": "object",
    "title": "Single Validator JSON schema",
    "additionalProperties": False,
    "required": ["vtype", "config"],
    "properties": {
        "vtype": {"type": "string", "enum": PropertyVTypes.values()},
        "config": {
            "type": "object",
            "properties": {
                "on_valid": {"type": "array", "items": effects_jsonschema_stub},
                "on_invalid": {"type": "array", "items": effects_jsonschema_stub},
            },
        },
    },
}

validators_options_json_schema = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "title": "Validators JSON schema",
    "type": "array",
    "items": validator_jsonschema_stub,
}

fileextensionvalidator_schema = copy.deepcopy(validator_jsonschema_stub)
fileextensionvalidator_schema["properties"]["config"]["properties"].update(
    {"accept": {"type": "array", "items": {"type": ["string", "null"]}}}
)

filesizevalidator_schema = copy.deepcopy(validator_jsonschema_stub)
filesizevalidator_schema["properties"]["config"]["properties"].update(
    {"maxSize": {"type": "number", "minimum": 0, "multipleOf": 1.0}}
)

#
# Representation of a property configurations
#

representation_jsonschema_stub = {
    "type": "object",
    "additionalProperties": False,
    "required": ["rtype", "config"],
    "properties": {
        "rtype": {"type": "string", "enum": _AllRepresentations.values()},
        "config": {"type": "object"},
    },
}

representation_options_json_schema = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "title": "Validators JSON schema",
    "type": "array",
    "items": representation_jsonschema_stub,
}

#
# Property Value Options json schema
#

options_json_schema = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "title": "Toplevel Property Options JSON schema",
    "type": "object",
    "additionalProperties": True,
    "properties": {
        # Validators
        "validators": validators_options_json_schema,
        # Reference Property additional options to hide columns and store filters
        "propmodels_excl": {"type": "array"},
        "propmodels_incl": {"type": "array"},
        "prefilters": {"type": "object"},
        # Representations
        "representation": representation_options_json_schema,
    },
}


#
# Scope Project Info json schema
#
# this is integregal copy of the definition in the KE-chain backend.
project_info_property_types = [
    PropertyType.CHAR_VALUE,
    PropertyType.TEXT_VALUE,
    PropertyType.INT_VALUE,
    PropertyType.FLOAT_VALUE,
    PropertyType.LINK_VALUE,
    PropertyType.DATETIME_VALUE,
    PropertyType.DATE_VALUE,
    PropertyType.TIME_VALUE,
    PropertyType.DURATION_VALUE,
]

scope_project_info_jsonschema = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "title": "Toplevel Scope Project info JSON schema",
    "description": "The project information properties that are part of the scope",
    "type": "array",
    "items": {
        "type": "object",
        "additionalProperties": False,
        "required": ["id", "name", "property_type", "value"],
        "properties": {
            "id": {"type": "integer", "minimum": 0},
            "name": {
                "type": ["string", "null"],
            },
            "property_type": {
                "type": "string",
                "enum": project_info_property_types,
                "default": PropertyType.CHAR_VALUE,
            },
            "value": {},
            "unit": {
                "type": ["string", "null"],
            },
            "description": {
                "type": ["string", "null"],
            },
        },
    },
}

#
# Prefill parts Info json schema
#
# this is integregal copy of the definition in the KE-chain backend.
prefill_property_stub = {
    "type": "object",
    "additionalProperties": False,
    "required": ["property_id", "value"],
    "properties": {
        "property_id": {"$ref": "#/definitions/uuidString"},
        "value": {},
    },
}
prefill_part_stub = {
    "type": "object",
    "additionalProperties": False,
    "required": ["name", "part_properties"],
    "properties": {
        "name": {"type": "string"},
        "part_properties": {"type": "array", "items": prefill_property_stub},
    },
}
form_collection_prefill_parts_schema = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "title": "FormCollection Prefill Parts schema",
    "version": "1.0.0",
    "type": "object",
    "additionalProperties": True,
    # propertyNames is a schema that all of an object's properties must be valid against
    # ref: https://json-schema.org/understanding-json-schema/reference/object.html#property-names
    "propertyNames": {"pattern": UUID_REGEX_PATTERN},
    # patternProperties is a schema that all of an objects property (named) must be valid against
    # ref: https://json-schema.org/understanding-json-schema/reference/object.html#pattern-properties
    "patternProperties": {
        UUID_REGEX_PATTERN: {
            "type": "array",
            "items": prefill_part_stub,
        }
    },
    "definitions": {
        "uuidString": {
            "type": "string",
            "pattern": UUID_REGEX_PATTERN,
        },
        "nullUuidString": {
            "type": ["string", "null"],
            "pattern": UUID_REGEX_PATTERN,
        },
        "nullString": {"type": ["string", "null"]},
        "positiveInteger": {"type": ["integer", "null"], "minimum": 0},
        "booleanNull": {"type": ["boolean", "null"]},
    },
}



from datetime import datetime
from typing import List, Optional, Union

import requests

from pykechain.enums import ContextGroup, ContextType
from pykechain.exceptions import APIError
from pykechain.models import BaseInScope
from pykechain.models.base import CrudActionsMixin, NameDescriptionTranslationMixin
from pykechain.models.input_checks import (
    check_base,
    check_datetime,
    check_enum,
    check_list_of_base,
    check_list_of_text,
    check_text,
    check_type,
)
from pykechain.models.tags import TagsMixin
from pykechain.typing import ObjectID, ObjectIDs
from pykechain.utils import Empty, clean_empty_values, empty, parse_datetime


class Context(
    BaseInScope, CrudActionsMixin, TagsMixin, NameDescriptionTranslationMixin
):
    """
    A virtual object representing a KE-chain Context.

    .. versionadded:: 3.12
    """

    url_detail_name = "context"
    url_list_name = "contexts"
    url_pk_name = "context_id"

    def __init__(self, json, **kwargs):
        """Construct a service from provided json data."""
        super().__init__(json, **kwargs)

        self.ref: str = json.get("ref")
        self.description: str = json.get("description", "")
        self.context_type: ContextType = json.get("context_type")
        self.options: dict = json.get("options", dict())
        self.context_group: ContextGroup = json.get("context_group", "")
        self._tags = json.get("tags")

        # associated activities
        self.activities: ObjectIDs = json.get("activities")

        # associated forms
        self.forms: ObjectIDs = json.get("form_collections")

        # LocationContext
        # these attributes are enabled when context_type == STATIC_LOCATION
        self.feature_collection: dict = json.get("feature_collection", dict())

        # TimeperiodContext
        # these attributes are enabled when context_type == TIME_PERIOD
        self.start_date: Optional[datetime] = parse_datetime(json.get("start_date"))
        self.due_date: Optional[datetime] = parse_datetime(json.get("due_date"))

    @classmethod
    def list(cls, client: "Client", **kwargs) -> List["Context"]:
        """List Context objects."""
        return super().list(client=client, **kwargs)

    @classmethod
    def get(cls, client: "Client", **kwargs) -> "Context":
        """Retrieve a single Context object."""
        return super().get(client=client, **kwargs)

    def edit(
        self,
        name: Optional[Union[str, Empty]] = empty,
        description: Optional[Union[str, Empty]] = empty,
        tags: Optional[List[Union[str, Empty]]] = empty,
        context_group: Optional[Union[ContextGroup, Empty]] = empty,
        scope: Optional[Union["Scope", ObjectID]] = empty,
        options: Optional[dict] = empty,
        activities: Optional[Union[List["Activity"], ObjectIDs]] = empty,
        feature_collection: Optional[dict] = empty,
        start_date: Optional[datetime] = empty,
        due_date: Optional[datetime] = empty,
        **kwargs,
    ) -> "Context":
        """
        Edit the Context.

        :param name: Name of the Context to be displayed to the end-user.
        :param scope: Scope object or Scope Id where the Context is active on.
        :param description: (optional) description of the Context
        :param activities: (optional) associated list of Activity or activity object ID
        :param tags: (optional) tags
        :param context_group: (optional) a context context_group of the choices of `ContextGroup`
        :param options: (optional) dictionary with options.
        :param feature_collection: (optional) dict with a geojson feature collection to store for a STATIC_LOCATION
        :param start_date: (optional) start datetime for a TIME_PERIOD context
        :param due_date: (optional) start datetime for a TIME_PERIOD context
        :return: a created Context Object
        :return: The updated Context object
        """
        from pykechain.models import Scope
        from pykechain.models import Activity

        update_dict = {
            "name": check_text(name, "name"),
            "description": check_text(description, "description"),
            "scope": check_base(scope, Scope, "scope"),
            "tags": check_list_of_text(tags, "tags"),
            "context_group": check_enum(context_group, ContextGroup, "context_group"),
            "activities": check_list_of_base(activities, Activity, "activities"),
            "options": check_type(options, dict, "options"),
            "feature_collection": check_type(
                feature_collection, dict, "feature_collection"
            ),
            "start_date": check_datetime(start_date, "start_date"),
            "due_date": check_datetime(due_date, "due_date"),
        }

        if feature_collection is None:
            update_dict["feature_collection"] = {}

        if kwargs:  # pragma: no cover
            update_dict.update(kwargs)

        update_dict = clean_empty_values(update_dict=update_dict)

        url = self._client._build_url("context", context_id=self.id)

        response = self._client._request("PUT", url, json=update_dict)

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError(f"Could not update Context: {self}", response=response)

        return self.refresh(json=response.json().get("results")[0])

    def link_activities(
        self, activities: Optional[List[Union["Activity", ObjectIDs]]] = empty, **kwargs
    ):
        """
        Link a context to one or more activities.

        :param activities: optional list of Activities or object Id's from activities.
        :returns: updated context objects
        """
        from pykechain.models import Activity

        update_dict = {
            "activities": check_list_of_base(activities, Activity, "activities"),
        }
        if kwargs:  # pragma: no cover
            update_dict.update(kwargs)

        url = self._client._build_url("context_link_activities", context_id=self.id)
        response = self._client._request("POST", url, json=update_dict)

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError(f"Could not update Context: {self}", response=response)

        return self.refresh(json=response.json().get("results")[0])

    def unlink_activities(
        self, activities: Optional[List[Union["Activity", ObjectIDs]]] = empty, **kwargs
    ):
        """
        Unlink a context to one or more activities.

        :param activities: optional list of Activities or object Id's from activities.
        :returns: updated context objects
        """
        from pykechain.models import Activity

        update_dict = {
            "activities": check_list_of_base(activities, Activity, "activities"),
        }
        if kwargs:  # pragma: no cover
            update_dict.update(kwargs)

        url = self._client._build_url("context_unlink_activities", context_id=self.id)
        response = self._client._request("POST", url, json=update_dict)

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError(f"Could not update Context: {self}", response=response)

        return self.refresh(json=response.json().get("results")[0])

    def delete(self):
        """Delete the Context."""
        return self._client.delete_context(self)



import warnings
from typing import Dict, List, Optional

import requests

from pykechain.defaults import API_EXTRA_PARAMS
from pykechain.exceptions import NotFoundError
from pykechain.models.input_checks import check_uuid, check_client
from pykechain.typing import ObjectID
from pykechain.utils import parse_datetime


class TooManyArgumentsWarning(UserWarning):
    """Warn a developer that too many arguments are used in the method call."""

    pass


class Base:
    """Base model connecting retrieved data to a KE-chain client.

    :ivar id: The UUID of the object (corresponds with the UUID in KE-chain).
    :type id: str
    :ivar name: The name of the object.
    :type name: basestring
    :ivar created_at: the datetime when the object was created if available (otherwise None)
    :type created_at: datetime or None
    :ivar updated_at: the datetime when the object was last updated if available (otherwise None)
    :type updated_at: datetime or None
    """

    def __init__(self, json: Dict, client: "Client"):
        """Construct a model from provided json data."""
        self._json_data = json
        self._client: "Client" = check_client(client)

        self.id = json.get("id")
        self.name = json.get("name")
        self.ref = json.get("ref")
        self.created_at = parse_datetime(json.get("created_at"))
        self.updated_at = parse_datetime(json.get("updated_at"))

    def __repr__(self):  # pragma: no cover
        return f"<pyke {self.__class__.__name__} '{self.name}' id {self.id[-8:]}>"

    def __eq__(self, other):  # pragma: no cover
        if hasattr(self, "id") and hasattr(other, "id"):
            return self.id == other.id
        else:
            return super().__eq__(other)

    def __hash__(self):
        return hash(self.id)

    def refresh(
        self,
        json: Optional[Dict] = None,
        url: Optional[str] = None,
        extra_params: Optional[Dict] = None,
    ):
        """Refresh the object in place.

        Can be called on the object without any arguments and should refresh the object inplace. If you want to
        use it in an advance way, you may call it with a json response from the server or provide the url to
        refetch the object from the server if the url cant be determined from the object itself.

        It is using the `Client.reload()` function to re-retrieve the object in a backend API call.

        :param json: (optional) json dictionary from a response from the server, will re-init object
        :type json: None or dict
        :param url: (optional) url to retrieve the object again, typically an identity url api/<service>/id
        :type url: None or basestring
        :param extra_params: (optional) additional paramenters (query params) for the request eg dict(fields='__all__')
        :type extra_params: None or dict
        """
        if json and isinstance(json, dict):
            self.__init__(json=json, client=self._client)
        else:
            src = self._client.reload(self, url=url, extra_params=extra_params)
            self.__dict__.update(src.__dict__)


class CrudActionsMixin:
    """
    Mixin that implements a list and get on the model.

    :cvar url_list_name: name of the list url in the defaults to contruct the api url
    :cvar url_detail_name: name of the detail url in the defaults to contruct the api url
    :cvar url_pk_name: name of the `<object>_id` as defined in the `API_PATH` in defaults
        for the detail url lookup on id.
    """

    url_list_name: str = None
    url_detail_name: str = None
    url_pk_name: str = None

    @classmethod
    def list(cls, client: "Client", **kwargs) -> List["self"]:
        """Retrieve a list of objects through the client."""
        if not cls.url_list_name:
            raise NotImplementedError(
                "This object type does not implement the list and get function on the object "
                "itself. It might be implemented in the Client object. \n"
                "[Pykechain Devs: If it is desired to implement the list/get functions "
                "on the object, the `cls.url_list_name` should match a uri string in the "
                "`defaults` for the `API_EXTRA_PARAMS` and the `API_PATH`.]"
            )

        kwargs.update(API_EXTRA_PARAMS[cls.url_list_name])
        response = client._request(
            "GET", client._build_url(cls.url_list_name), params=kwargs
        )

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise NotFoundError(f"Could not retrieve {cls.__name__}", response=response)

        return [cls(json=j, client=client) for j in response.json()["results"]]

    @classmethod
    def get(cls, client: "Client", **kwargs) -> "self":
        """Retrieve a single object using the client."""
        pk = None
        if "pk" in kwargs:
            pk = check_uuid(kwargs.pop("pk"))
        elif "id" in kwargs:
            # for accidental use of 'id' on the obj retrieving we have this smart
            # popper from the kwargs to ensure that the intention is still correct
            # it will result in a single obj retrieve.
            pk = check_uuid(kwargs.pop("id"))

        if pk:
            if not cls.url_detail_name:
                raise NotImplementedError(
                    "This object type does not implement the list and get function on the object "
                    "itself. It might be implemented in the Client object. \n"
                    "[Pykechain Devs: If it is desired to implement the list/get functions "
                    "on the object, the `cls.url_detail_name` should match a uri string in the "
                    "`defaults` for the `API_EXTRA_PARAMS` and the `API_PATH`.]"
                )
            # if more kwargs are provided, warn the developer that these will be ignored
            # and that the dev should alter the call to provide the `pk` only
            if kwargs:
                warnings.warn(
                    f"Too many arguments are passed to the method. Only the `pk` (or `id`) "
                    f"argument is used. The other arguments are not passed to the API and "
                    f"have no effect. Please alter the call to this method to reduce the "
                    f"number of arguments or use the '`list()` of <object>s' equivalent. "
                    f"Eg. if you want to filter on `pk` AND `category` you can use the "
                    f"list function. Got: {kwargs}",
                    UserWarning,
                )

            field_name_in_api_path = cls.url_pk_name or f"{cls.__name__.lower()}_id"
            url = client._build_url(cls.url_detail_name, **{field_name_in_api_path: pk})

            request_params = {}
            if cls.url_detail_name in API_EXTRA_PARAMS:
                request_params = API_EXTRA_PARAMS.get(cls.url_detail_name)
            elif cls.url_list_name in API_EXTRA_PARAMS:
                request_params = API_EXTRA_PARAMS.get(cls.url_list_name)

            response = client._request("GET", url, params=request_params)
            if response.status_code != requests.codes.ok:  # pragma: no cover
                raise NotFoundError("Could not retrieve object", response=response)
            return cls(response.json()["results"][0], client=client)

        # otherwise do the normal singular retrieve
        return client._retrieve_singular(cls.list, client=client, **kwargs)

    def delete(self) -> None:
        """
        Delete the object.

        After successful deletion on the server it will set the id to None to give feedback to the
        user that there is no relation between the object in memery and an object in KE-chain.
        """
        field_name_in_api_path = self.url_pk_name or f"{self.__name__.lower()}_id"
        url = self._client._build_url(
            self.url_detail_name, **{field_name_in_api_path: self.id}
        )

        response = self._client._request("DELETE", url)
        if response.status_code != requests.codes.no_content:
            raise NotFoundError("Could not delete object", response=response)
        # reset the id to None to feedback that the object is deleted in KE-chain
        self.id = None
        return None


class BaseInScope(Base):
    """
    Base model for KE-chain objects coupled to a scope.

    :ivar scope_id: UUID of the Scope
    :type scope_id: str
    """

    def __init__(self, json, *args, **kwargs):
        """Append the scope ID to the attributes of the base object."""
        super().__init__(json, *args, **kwargs)

        self.scope_id: Optional[ObjectID] = json.get(
            "scope_id", json.get("scope", None)
        )
        self._scope: Optional["Scope"] = None

    @property
    def scope(self):
        """
        Scope this object belongs to.

        This property will return a `Scope` object. It will make an additional call to the KE-chain API.

        :return: the scope
        :type: :class:`pykechain.models.Scope`
        :raises NotFoundError: if the scope could not be found
        """
        if not self._scope and self.scope_id:
            self._scope = self._client.scope(pk=self.scope_id, status=None)
        return self._scope


class NameDescriptionTranslationMixin:
    """Mixin that includes translations of the name and description of an object."""

    pass



import uuid

import jsonschema

from pykechain.enums import (
    FormCategory,
    Multiplicity,
    PropertyType,
    StatusCategory,
    WorkflowCategory,
)
from pykechain.exceptions import (
    APIError,
    ForbiddenError,
    IllegalArgumentError,
    NotFoundError,
)
from pykechain.models.form import Form
from tests.classes import TestBetamax


class TestForms(TestBetamax):
    """
    Test retrieval and forms attributes and methods.
    """

    def setUp(self):
        super().setUp()
        self.cross_scope_project = self.client.scope(ref="cannondale-project")
        self.workflow = self.client.workflow(
            name="Simple Form Flow", category=WorkflowCategory.CATALOG
        )
        self.discipline_context = self.project.context(name="Discipline 1")
        self.asset_context = self.project.context(name="Object 1")
        self.form_model_name = "__TEST__FORM_MODEL"
        self.form_model_name_2 = "__TEST__FORM_MODEL_2"

        self.form_model = self.client.create_form_model(
            name=self.form_model_name,
            scope=self.project,
            workflow=self.workflow,
            category=FormCategory.MODEL,
            contexts=[self.asset_context, self.discipline_context],
        )
        self.form_model_2 = self.project.create_form_model(
            name=self.form_model_name_2,
            workflow=self.workflow,
            category=FormCategory.MODEL,
            contexts=[self.asset_context],
        )
        self.form_instance = self.form_model.instantiate(name="__TEST_FORM_INSTANCE")
        self.test_form_instance = None
        self.cloned_form_model = None
        self.form_model_for_deletion = None
        self.cloned_workflow = None

    def tearDown(self):
        super().tearDown()
        if self.form_instance:
            try:
                self.form_instance.delete()
            except (NotFoundError, APIError):
                pass
        if self.test_form_instance:
            try:
                self.test_form_instance.delete()
            except (NotFoundError, APIError):
                pass
        if self.form_model:
            try:
                self.form_model.delete()
            except (NotFoundError, APIError):
                pass
        if self.form_model_2:
            try:
                self.form_model_2.delete()
            except (NotFoundError, APIError):
                pass
        if self.form_model_for_deletion:
            try:
                self.form_model_for_deletion.delete()
            except (NotFoundError, APIError):
                pass
        if self.cloned_form_model:
            try:
                self.cloned_form_model.delete()
            except (NotFoundError, APIError):
                pass
        if self.cloned_workflow:
            try:
                self.cloned_workflow.delete()
            except (NotFoundError, APIError):
                pass

    def test_form_attributes(self):
        attributes = [
            "_client",
            "_json_data",
            "id",
            "name",
            "created_at",
            "updated_at",
            "derived_from_id",
            "ref",
            "description",
            "active_status",
            "category",
            "tags",
            "form_model_root",
            "form_instance_root",
            "model_id",
            "scope_id",
            "status_forms",
        ]

        for attr in attributes:
            with self.subTest(attr):
                self.assertTrue(
                    hasattr(self.form_model, attr),
                    f"Could not find '{attr}' in the form: '{self.form_model.__dict__.keys()}",
                )

    def test_create(self):
        self.assertIsInstance(self.form_model, Form)
        self.assertEqual(self.form_model.category, FormCategory.MODEL)

        self.assertIsInstance(self.form_model_2, Form)
        self.assertEqual(self.form_model_2.category, FormCategory.MODEL)

        self.assertIsInstance(self.form_instance, Form)
        self.assertEqual(self.form_instance.category, FormCategory.INSTANCE)

    def test_form_instances(self):
        with self.subTest("testing instances"):
            form_instances = self.form_model.instances()
            self.assertTrue(len(form_instances) >= 1)
            form_instance = form_instances[0]

            self.assertTrue(form_instance.category, FormCategory.INSTANCE)
            self.assertTrue(form_instance.model_id, self.form_model.id)

        with self.subTest("testing instance"):
            form_instance = self.form_model.instances()[0]
            self.assertTrue(form_instance.category, FormCategory.INSTANCE)
            self.assertTrue(form_instance.model_id, self.form_model.id)

    def test_model_form_instantiation(self):
        self.test_form_instance = self.form_model.instantiate(
            name="__TEST_FORM_INSTANCE"
        )

        self.assertTrue(self.test_form_instance.model_id, self.form_model.id)

    def test_model_form_instantiation_from_client(self):
        self.test_form_instance = self.client.instantiate_form(
            name="___TEST_FORM_INSTANCE", model=self.form_model
        )
        self.assertIsInstance(self.test_form_instance, Form)
        self.assertEqual(self.test_form_instance.category, FormCategory.INSTANCE)
        self.assertTrue(self.test_form_instance.model_id, self.form_model.id)

    def test_model_form_instantiation_from_scope(self):
        self.test_form_instance = self.project.instantiate_form(
            name="___TEST_FORM_INSTANCE", model=self.form_model
        )
        self.assertIsInstance(self.test_form_instance, Form)
        self.assertEqual(self.test_form_instance.category, FormCategory.INSTANCE)
        self.assertTrue(self.test_form_instance.model_id, self.form_model.id)

    def test_model_form_instantiation_from_wrong_scope(self):
        with self.assertRaises(IllegalArgumentError):
            self.test_form_instance = self.cross_scope_project.instantiate_form(
                name="___TEST_FORM_INSTANCE", model=self.form_model
            )

    def test_model_clone_same_scope(self):
        self.cloned_form_model = self.form_model.clone(
            name="__TEST_FORM_MODEL_CLONE",
            target_scope=self.project,
            contexts=[self.asset_context, self.discipline_context],
        )
        self.assertTrue(self.cloned_form_model)
        self.assertIsInstance(self.cloned_form_model, Form)
        self.assertEqual(self.form_model.scope_id, self.cloned_form_model.scope_id)
        self.assertIn(
            self.asset_context.id,
            [
                context["id"]
                for context in self.cloned_form_model._json_data["contexts"]
            ],
        )
        self.assertIn(
            self.discipline_context.id,
            [
                context["id"]
                for context in self.cloned_form_model._json_data["contexts"]
            ],
        )

    def test_model_clone_no_target_scope(self):
        self.cloned_form_model = self.form_model.clone(
            name="__TEST_FORM_MODEL_CLONE",
            contexts=[self.asset_context, self.discipline_context],
        )
        self.assertTrue(self.cloned_form_model)
        self.assertIsInstance(self.cloned_form_model, Form)
        self.assertEqual(self.form_model.scope_id, self.cloned_form_model.scope_id)
        self.assertIn(
            self.asset_context.id,
            [
                context["id"]
                for context in self.cloned_form_model._json_data["contexts"]
            ],
        )
        self.assertIn(
            self.discipline_context.id,
            [
                context["id"]
                for context in self.cloned_form_model._json_data["contexts"]
            ],
        )

    def test_model_clone_cross_scope(self):
        self.cloned_form_model = self.form_model.clone(
            name="__TEST_FORM_MODEL_CLONE_CROSS_SCOPE",
            target_scope=self.cross_scope_project,
            contexts=[self.asset_context, self.discipline_context],
        )
        self.assertTrue(self.cloned_form_model)
        self.assertIsInstance(self.cloned_form_model, Form)
        self.assertEqual(self.cross_scope_project.id, self.cloned_form_model.scope_id)
        self.assertIn(
            self.asset_context.name,
            [
                context["name"]
                for context in self.cloned_form_model._json_data["contexts"]
            ],
        )
        self.assertIn(
            self.discipline_context.name,
            [
                context["name"]
                for context in self.cloned_form_model._json_data["contexts"]
            ],
        )

    def test_form_retrieve_by_name_from_scope(self):
        form = self.project.form(name=self.form_model_name)

        self.assertIsInstance(form, Form)

    def test_forms_retrieve_by_context_from_scope(self):
        forms = self.project.forms(context=[self.asset_context])

        self.assertEqual(len(forms), 2)
        for form in forms:
            self.assertIsInstance(form, Form)
        self.assertIn(self.form_model.id, [form.id for form in forms])
        self.assertIn(self.form_model_2.id, [form.id for form in forms])

    def test_forms_retrieve_instances_by_model_from_scope(self):
        form_instances = self.project.forms(
            model=self.form_model, category=FormCategory.INSTANCE
        )

        self.assertTrue(len(form_instances) == 1)
        form_instance = form_instances[0]

        self.assertTrue(form_instance.category, FormCategory.INSTANCE)
        self.assertTrue(form_instance.model_id, self.form_model.id)

    def test_form_model_edit(self):
        new_name = "__TEST_EDITED_FORM_MODEL"
        self.form_model_2.edit(name=new_name)
        self.assertEqual(self.form_model_2.name, new_name)

    def test_form_model_edit_with_instances(self):
        new_name = "__TEST_EDITED_FORM_MODEL"
        self.form_model.edit(name=new_name)
        self.assertEqual(self.form_model.name, new_name)

    def test_form_instance_edit(self):
        new_name = "__TEST_INSTANCE_EDITED"
        self.form_instance.edit(name=new_name)
        self.assertEqual(self.form_instance.name, new_name)

    def test_forms_delete(self):
        form_model_name = "__TEST_FORM_MODEL_FOR_DELETION"
        self.form_model_for_deletion = self.client.create_form_model(
            name=form_model_name,
            scope=self.project,
            workflow=self.workflow,
            category=FormCategory.MODEL,
            contexts=[self.asset_context, self.discipline_context],
        )
        self.assertIsInstance(self.form_model_for_deletion, Form)
        self.form_model_for_deletion.delete()
        with self.assertRaises(APIError, msg="Cant delete the same Form twice!"):
            self.form_model_for_deletion.delete()
        with self.assertRaises(NotFoundError, msg="Deleted Form cannot be found!"):
            self.project.form(name=form_model_name)

    def test_form_has_part_is_true(self):
        part_model = self.project.model(name=self.form_model.name)
        has_part_check = self.form_model.has_part(part=part_model)
        self.assertTrue(has_part_check)

        part_instance = part_model.instance()
        has_part_check = self.form_instance.has_part(part=part_instance)
        self.assertTrue(has_part_check)

    def test_form_has_part_is_false(self):
        part_model = self.project.model(name="Bike")

        has_part_check = self.form_model.has_part(part=part_model)
        self.assertFalse(has_part_check)

    def test_form_has_part_wrong_inputs(self):
        with self.assertRaises(IllegalArgumentError):
            self.form_model.has_part(part=self.discipline_context)

    def test_form_workflow_compatible_within_scope(self):
        workflows = self.form_model.workflows_compatible_with_scope(scope=self.project)
        self.assertEqual(len(workflows), 1)
        self.cloned_workflow = self.workflow.clone(
            target_scope=self.project, name="CLONED WORKFLOW"
        )
        workflows = self.form_model.workflows_compatible_with_scope(scope=self.project)
        self.assertEqual(len(workflows), 2)

    def test_form_workflow_compatible_within_scope_wrong_inputs(self):
        with self.assertRaises(IllegalArgumentError):
            self.form_model.workflows_compatible_with_scope(
                scope=self.discipline_context
            )


class TestFormsBulk(TestBetamax):
    """
    Test bulk forms attributes and methods.
    """

    def setUp(self):
        super().setUp()
        self.workflow = self.client.workflow(
            name="Simple Form Flow", category=WorkflowCategory.CATALOG
        )
        self.discipline_context = self.project.context(name="Discipline 1")
        self.asset_context = self.project.context(name="Object 1")
        self.form_model_name = "__TEST__FORM_MODEL"
        self.form_model = self.client.create_form_model(
            name=self.form_model_name,
            scope=self.project,
            workflow=self.workflow,
            category=FormCategory.MODEL,
            contexts=[self.asset_context, self.discipline_context],
        )
        self.new_forms = list()
        for idx in range(1, 5):
            form_dict = {
                "form": self.form_model,
                "values": {
                    "name": f"Form {idx}",
                    "contexts": [self.asset_context],
                },
            }
            self.new_forms.append(form_dict)
        self.forms_created = None

    def tearDown(self):
        super().tearDown()
        if self.forms_created:
            for form in self.forms_created:
                try:
                    form.delete()
                except APIError:
                    pass
        if self.form_model:
            try:
                self.form_model.delete()
            except (ForbiddenError, APIError):
                pass

    def test_bulk_instantiate_forms(self):
        self.forms_created = self.client._create_forms_bulk(
            forms=self.new_forms, retrieve_instances=True
        )
        for form in self.forms_created:
            self.assertIsInstance(form, Form)
            self.assertEqual(form.category, FormCategory.INSTANCE)
            self.assertEqual(form.model_id, self.form_model.id)

    def test_bulk_delete_forms(self):
        self.forms_created = self.client._create_forms_bulk(
            forms=self.new_forms, retrieve_instances=True
        )
        input_forms_and_uuids = [
            self.forms_created[0],
            self.forms_created[1],
            self.forms_created[2].id,
            self.forms_created[3].id,
        ]
        self.client._delete_forms_bulk(forms=input_forms_and_uuids)
        for idx in range(1, 5):
            with self.subTest(idx=idx):
                with self.assertRaises(NotFoundError):
                    self.project.form(name=f"Form {idx}")

    def test_bulk_delete_forms_with_wrong_input(self):
        wrong_input = [self.project.activity(name="Specify wheel diameter")]
        with self.assertRaises(IllegalArgumentError):
            self.client._delete_forms_bulk(forms=wrong_input)


class TestFormsMethods(TestBetamax):
    """
    Test linking and unlinking contexts to forms.
    """

    def setUp(self):
        super().setUp()
        self.workflow = self.client.workflow(
            name="Simple Form Flow", category=WorkflowCategory.CATALOG
        )
        self.discipline_context = self.project.context(name="Discipline 1")
        self.asset_context = self.project.context(name="Object 1")
        self.location_context = self.project.context(name="Location 1")
        self.form_model_name = "__TEST__FORM_MODEL"
        self.form_model = self.client.create_form_model(
            name=self.form_model_name,
            scope=self.project,
            workflow=self.workflow,
            category=FormCategory.MODEL,
            contexts=[self.asset_context],
        )
        self.form_instance_to_link = self.form_model.instantiate(
            name="__TEST_FORM_INSTANCE_LINK"
        )
        self.form_instance_to_unlink = self.form_model.instantiate(
            name="__TEST_FORM_INSTANCE_UNLINK",
            contexts=[
                self.asset_context,
                self.discipline_context,
                self.location_context,
            ],
        )
        self.form_instance_to_set_assignees = self.form_model.instantiate(
            name="__TEST_FORM_INSTANCE_SET_STATUS_ASSIGNEES"
        )
        self.form_instance_to_transition = self.form_model.instantiate(
            name="__TEST_FORM_INSTANCE_APPLY_TRANSITION"
        )

    def tearDown(self):
        super().tearDown()
        if self.form_instance_to_link:
            try:
                self.form_instance_to_link.delete()
            except (ForbiddenError, APIError):
                pass
        if self.form_instance_to_unlink:
            try:
                self.form_instance_to_unlink.delete()
            except (ForbiddenError, APIError):
                pass
        if self.form_instance_to_set_assignees:
            try:
                self.form_instance_to_set_assignees.delete()
            except (ForbiddenError, APIError):
                pass
        if self.form_instance_to_transition:
            try:
                self.form_instance_to_transition.delete()
            except (ForbiddenError, APIError):
                pass
        if self.form_model:
            try:
                self.form_model.delete()
            except (ForbiddenError, APIError):
                pass

    def test_link_contexts_to_form_model(self):
        self.form_model.link_contexts(
            contexts=[self.discipline_context, self.location_context]
        )
        self.assertIn(
            self.asset_context.id,
            [context["id"] for context in self.form_model._json_data["contexts"]],
        )
        self.assertIn(
            self.discipline_context.id,
            [context["id"] for context in self.form_model._json_data["contexts"]],
        )
        self.assertIn(
            self.location_context.id,
            [context["id"] for context in self.form_model._json_data["contexts"]],
        )
        self.assertEqual(len(self.form_model._json_data["contexts"]), 3)

    def test_link_contexts_to_form_instance(self):
        self.form_instance_to_link.link_contexts(contexts=[self.location_context])
        self.assertIn(
            self.location_context.id,
            [
                context["id"]
                for context in self.form_instance_to_link._json_data["contexts"]
            ],
        )
        self.assertEqual(len(self.form_instance_to_link._json_data["contexts"]), 1)

    def test_unlink_contexts_to_form_model(self):
        self.form_model.unlink_contexts(contexts=[self.asset_context])
        self.assertEqual(len(self.form_model._json_data["contexts"]), 0)

    def test_unlink_contexts_to_form_instance(self):
        self.form_instance_to_unlink.unlink_contexts(
            contexts=[self.location_context, self.asset_context]
        )
        self.assertIn(
            self.discipline_context.id,
            [
                context["id"]
                for context in self.form_instance_to_unlink._json_data["contexts"]
            ],
        )
        self.assertEqual(len(self.form_instance_to_unlink._json_data["contexts"]), 1)

    def test_unlink_non_connected_contexts_from_form(self):
        with self.assertRaises(APIError):
            self.form_model.unlink_contexts(
                contexts=[self.discipline_context, self.asset_context]
            )

    def test_set_status_assignees(self):
        status_assignees_list = list()
        test_manager = self.client.user("testmanager")
        supervisor = self.client.user("supervisor")
        test_lead = self.client.user("testlead")
        for status_form in self.form_instance_to_set_assignees.status_forms:
            status_dict = {
                "status": status_form.status,
                "assignees": [test_manager, supervisor, test_lead],
            }
            status_assignees_list.append(status_dict)
        self.form_instance_to_set_assignees.set_status_assignees(
            statuses=status_assignees_list
        )
        for status_form in self.form_instance_to_set_assignees._status_assignees:
            self.assertIn(
                test_manager.id,
                [status_assignee["id"] for status_assignee in status_form["assignees"]],
            )
            self.assertEqual(len(status_form["assignees"]), 3)

    def test_set_status_assignees_with_wrong_format(self):
        status_assignees_list = list()
        test_manager = self.client.user("testmanager")
        for status_form in self.form_instance_to_set_assignees.status_forms:
            status_dict = {"id": status_form.status, "assignees": [test_manager]}
            status_assignees_list.append(status_dict)
        with self.assertRaises(IllegalArgumentError):
            self.form_instance_to_set_assignees.set_status_assignees(
                statuses=status_assignees_list
            )

    def test_set_status_assignees_with_wrong_input(self):
        status_assignees_list = list()
        test_manager = self.client.user("testmanager")
        for status_form in self.form_instance_to_set_assignees.status_forms:
            status_dict = {
                "status": status_form.id,  # This should be status_form.status
                "assignees": [test_manager],
            }
            status_assignees_list.append(status_dict)
        with self.assertRaises(APIError):
            self.form_instance_to_set_assignees.set_status_assignees(
                statuses=status_assignees_list
            )

    def test_apply_transition(self):
        in_progress_transition = self.workflow.transition("In progress")
        done_transition = self.workflow.transition("Done")
        self.assertEqual(
            self.form_instance_to_transition.active_status["status_category"],
            StatusCategory.TODO,
        )
        self.form_instance_to_transition.apply_transition(
            transition=in_progress_transition
        )
        self.assertEqual(
            self.form_instance_to_transition.active_status["status_category"],
            StatusCategory.INPROGRESS,
        )
        self.form_instance_to_transition.apply_transition(transition=done_transition)
        self.assertEqual(
            self.form_instance_to_transition.active_status["status_category"],
            StatusCategory.DONE,
        )

    def test_apply_transition_with_wrong_input(self):
        with self.assertRaises(IllegalArgumentError):
            self.form_instance_to_transition.apply_transition(transition=self.workflow)

    def test_retrieve_possible_transitions(self):
        possible_transitions = self.form_instance_to_transition.possible_transitions()
        for transition in possible_transitions:
            self.assertIn(transition.id, [t.id for t in self.workflow.transitions])
        self.assertEqual(len(possible_transitions), 5)


class TestFormsPreFillPartMethods(TestBetamax):
    """
    Test linking and unlinking contexts to forms.
    """

    def setUp(self):
        super().setUp()
        self.workflow = self.client.workflow(
            name="Simple Form Flow", category=WorkflowCategory.CATALOG
        )
        self.form_model_name = "__TEST__FORM_MODEL"
        self.form_model = self.client.create_form_model(
            name=self.form_model_name,
            scope=self.project,
            workflow=self.workflow,
            category=FormCategory.MODEL,
            contexts=[],
        )
        self.form_model_root = self.form_model.form_model_root
        self.a_part_model = self.client.create_model(
            parent=self.form_model_root,
            name="A Part Model",
            multiplicity=Multiplicity.ZERO_MANY,
        )
        self.p1 = self.client.create_property(
            model=self.a_part_model, name="p1", property_type=PropertyType.CHAR_VALUE
        )

    def tearDown(self):
        super().tearDown()
        if self.form_model:
            try:
                self.form_model.delete()
            except (ForbiddenError, APIError):
                pass

    def test_form_prefill_parts_can_be_set(self):
        prefill_parts = {
            self.a_part_model.id: [
                dict(
                    name="1st instance",
                    part_properties=[dict(property_id=self.p1.id, value=1)],
                ),
                dict(
                    name="2nd instance",
                    part_properties=[dict(property_id=self.p1.id, value=2)]
                )
            ]
        }

        self.assertDictEqual(self.form_model._prefill_parts, {})
        self.form_model.set_prefill_parts(prefill_parts=prefill_parts)
        self.assertDictEqual(prefill_parts, self.form_model._prefill_parts)

    def test_form_prefill_parts_with_wrong_payload(self):
        prefill_parts = {"foo": "bar"}
        self.assertDictEqual(self.form_model._prefill_parts, {})
        with self.assertRaisesRegex(jsonschema.ValidationError, "'foo' does not match"):
            self.form_model.set_prefill_parts(prefill_parts=prefill_parts)

        prefill_parts = {str(uuid.uuid4()): ["bar"]}
        self.assertDictEqual(self.form_model._prefill_parts, {})
        with self.assertRaisesRegex(jsonschema.ValidationError, "'bar' is not of type 'object'"):
            self.form_model.set_prefill_parts(prefill_parts=prefill_parts)

        prefill_parts = {str(uuid.uuid4()): [{"name": "bar"}]}
        self.assertDictEqual(self.form_model._prefill_parts, {})
        with self.assertRaisesRegex(jsonschema.ValidationError, "'part_properties' is a required property"):
            self.form_model.set_prefill_parts(prefill_parts=prefill_parts)

        prefill_parts = {str(uuid.uuid4()): [{"name": "bar", "part_properties": {}}]}
        self.assertDictEqual(self.form_model._prefill_parts, {})
        with self.assertRaisesRegex(jsonschema.ValidationError, "{} is not of type 'array'"):
            self.form_model.set_prefill_parts(prefill_parts=prefill_parts)



from pykechain.models.widgets.widget import Widget
from pykechain.models.widgets.widget_schemas import undefined_meta_schema


# UNDEFINED
# PROPERTYGRID
# SUPERGRID
# HTML
# FILTEREDGRID
# SERVICE
# NOTEBOOK
# ATTACHMENTVIEWER
# TASKNAVIGATIONBAR
# JSON
# METAPANEL
# FORMMETAPANEL
# MULTICOLUMN
# SCOPE_WIDGET
# THIRD_PARTY
# PROGRESS
# SIGNATURE
# CARD =
# WEATHER
# DASHBOARD
# SCOPEMEMBERS
# PROJECTINFO

#
# The names of all the widgets do have a Pattern conforming to the following
#
# rule: "<widget_type_in_undercast_with_first_letter_capitalized>Widget"
# In regex terms: r"[A-Z][a-z]+Widget"
#

class MetapanelWidget(Widget):
    """Metapanel Widget."""


class FormmetapanelWidget(Widget):
    """FormMetapanel Widget."""


class PropertygridWidget(Widget):
    """Propertygrid Widget."""


class UndefinedWidget(Widget):
    """Undefined Widget."""

    schema = undefined_meta_schema


class FilteredgridWidget(Widget):
    """Filteredgrid Widget."""


class SupergridWidget(Widget):
    """Supergrid Widget."""


class AttachmentviewerWidget(Widget):
    """Attachmentviewer Widget."""


class TasknavigationbarWidget(Widget):
    """Tasknavigationbar Widget."""


class HtmlWidget(Widget):
    """HTML Widget."""


class ServiceWidget(Widget):
    """Service Widget."""


class NotebookWidget(Widget):
    """Notebook Widget."""


class JsonWidget(Widget):
    """JSON Widget."""


class MulticolumnWidget(Widget):
    """Multicolumn Widget."""


class ProgressWidget(Widget):
    """Progress bar Widget."""


class ScopeWidget(Widget):
    """Scope grid Widget."""


class SignatureWidget(Widget):
    """Signature Widget."""


class CardWidget(Widget):
    """Card Widget."""


class ThirdpartyWidget(Widget):
    """Thirdparty Widget."""


class TasksWidget(Widget):
    """Tasks Widget."""


class WeatherWidget(Widget):
    """Weather Widget."""


class ServicecardWidget(Widget):
    """ServiceCard Widget."""


class DashboardWidget(Widget):
    """Dashboard Widget."""


class ScopemembersWidget(Widget):
    """ScopeMembers Widget."""


class ProjectinfoWidget(Widget):
    """Project Info Widget."""



from typing import Dict, List, Optional, Union

import requests

from pykechain.defaults import API_EXTRA_PARAMS
from pykechain.enums import FormCategory
from pykechain.exceptions import APIError, ForbiddenError
from pykechain.models import Activity, Part, Scope
from pykechain.models.base import (
    Base,
    BaseInScope,
    CrudActionsMixin,
    NameDescriptionTranslationMixin,
)
from pykechain.models.context import Context
from pykechain.models.input_checks import (
    check_base,
    check_json,
    check_list_of_base,
    check_list_of_dicts,
    check_text,
)
from pykechain.models.tags import TagsMixin
from pykechain.models.validators.validator_schemas import (
    form_collection_prefill_parts_schema,
)
from pykechain.models.workflow import Status, Transition
from pykechain.typing import ObjectID
from pykechain.utils import Empty, clean_empty_values, empty


class StatusForm(Base):
    """A virtual object representing a KE-chain StatusForm.

    A StatusForm is an intermediate object linking the Forms to its 'subforms', where each
    status of a form has a link to its Activity.
    """

    def __init__(self, json, **kwargs):
        """Construct a status form from provided json data."""
        super().__init__(json, **kwargs)
        self.description: str = json.get("description", "")
        self.ref: str = json.get("ref", "")
        self.status: Status = Status(json.get("status"), client=self._client)
        self.activity: Activity = Activity(json.get("activity"), client=self._client)
        self.form: str = json.get("form")

    def __repr__(self):  # pragma: no cover
        return f"<pyke StatusForm  '{self.status}' id {self.id[-8:]}>"


class Form(BaseInScope, CrudActionsMixin, TagsMixin, NameDescriptionTranslationMixin):
    """
    A virtual object representing a KE-chain Form Collection.

    .. versionadded:: v3.20

    :ivar id: id of the form
    :ivar name: name of the form
    :ivar description: description of the form
    :ivar ref: reference (slugified name) of the form
    :ivar category: the category of the form (FormCategory.MODEL or FormCategory.INSTANCE)
    :ivar active_status: current active Status of the form
    :ivar form_model_root: the form model root Part
    :ivar form_instance_root: the form instance root Part
    :ivar model_id: the pk of the model of this form instance
    :ivar status_forms: The StatusForms of this form
    :ivar prefill_parts: The parts to prefill when a Form Model is isntantiated.
    :ivar created_at: datetime in UTC timezone when the form was created
    :ivar updated_at: datetime in UTC timezone when the form was last updated
    :ivar derived_from_id: (optional) id where the objects is derived from.
    """

    url_detail_name = "form"
    url_list_name = "forms"
    url_pk_name = "form_id"

    def __init__(self, json, **kwargs):
        """Construct a service from provided json data."""
        super().__init__(json, **kwargs)
        self.description: str = json.get("description", "")
        self.ref: str = json.get("ref", "")
        self._workflow = json.get("workflow")
        self.active_status: "Status" = json.get("active_status")
        self.form_model_root: "Part" = json.get("form_model_root")
        self.form_instance_root: "Part" = json.get("form_instance_root")
        self.model_id: "Form" = json.get("model")
        self._prefill_parts: dict = json.get("prefill_parts")
        self.derived_from_id: Optional[ObjectID] = json.get("derived_from_id")
        self.active: bool = json.get("active")
        self.category: FormCategory = json.get("category")
        self._status_forms: List[Dict] = json.get("status_forms", [])
        self._status_assignees: List[Dict] = json.get(
            "status_assignees_has_widgets", []
        )
        self.contexts = json.get("contexts", [])
        self.context_groups = json.get("context_groups", [])

    def __repr__(self):  # pragma: no cover
        return f"<pyke Form  '{self.name}' '{self.category}' id {self.id[-8:]}>"

    @classmethod
    def create_model(
        cls,
        client: "Client",
        name: str,
        scope: Union[Scope, ObjectID],
        workflow: Union["Workflow", ObjectID],
        contexts: List[Union[Context, ObjectID]],
        **kwargs: dict,
    ) -> "Form":
        """Create a new form model.

        Needs scope, name, workflow.
        :param client: The client connection to KE-chain API
        :param name: Name of the new form model
        :param scope: Scope or scope_id where to create the form model
        :param workflow: Workflow or workflow_id of the workflow associated to the form model,
            should be in scope.
        :param contexts: A list of Context or context id's to link to the Form model.
        :param kwargs: Additional kwargs such as contexts.
        """
        from pykechain.models.workflow import Workflow

        data = {
            "name": check_text(name, "name"),
            "scope": check_base(scope, Scope, "scope"),
            "workflow": check_base(workflow, Workflow, "workflow"),
            "contexts": check_list_of_base(contexts, Context, "contexts"),
        }

        kwargs.update(API_EXTRA_PARAMS[cls.url_list_name])
        response = client._request(
            "POST", client._build_url(cls.url_list_name), params=kwargs, json=data
        )

        if response.status_code != requests.codes.created:  # pragma: no cover
            raise APIError(f"Could not create {cls.__name__}", response=response)

        return cls(json=response.json()["results"][0], client=client)

    #
    # Concepts underneath the Form
    # StatusForms

    @property
    def status_forms(self):
        """Retrieve the Status Forms of this Form."""
        return [StatusForm(s, client=self._client) for s in self._status_forms]

    #
    # @classmethod
    # def list(cls, client: "Client", **kwargs) -> List["Form"]:
    #     """Retrieve a list of Form objects through the client."""
    #     return super().list(client=client, **kwargs)
    #
    # @classmethod
    # def get(cls, client: "Client", **kwargs) -> "Form":
    #     """Retrieve a Form object through the client."""
    #     return super().get(client=client, **kwargs)

    @property
    def is_model(self) -> bool:
        """Form is a Form Model or Form Template."""
        return self.category == FormCategory.MODEL

    @property
    def is_instance(self) -> bool:
        """Form is a Form Instance."""
        return self.category == FormCategory.INSTANCE

    @property
    def is_active(self) -> bool:
        """Form is an active Form."""
        return self.active is not None and self.active

    #
    # Form Model finders and searchers.
    #

    def instances(self, **kwargs) -> [List["Form"]]:
        """Retrieve the instances of this Form Model."""
        if self.category == FormCategory.MODEL:
            return self._client.forms(
                model=self, category=FormCategory.INSTANCE, **kwargs
            )
        else:
            raise FileNotFoundError(
                f"Form '{self}' is not a model, hence it has no instances."
            )

    def instance(self, **kwargs) -> "Form":
        """Retrieve a single instance of a Form Model."""
        return self._client._retrieve_singular(self.instances, **kwargs)

    def edit(
        self,
        name: Optional[Union[str, Empty]] = empty,
        description: Optional[Union[str, Empty]] = empty,
        **kwargs,
    ) -> None:
        """Edit the details of a Form (model or instance).

        Setting an input to None will clear out the value (exception being name).

        :param name: optional name of the form to edit. Cannot be cleared.
        :type name: basestring or None or Empty
        :param description: (optional) description of the form. Can be cleared.
        :type description: basestring or None or Empty
        :param kwargs: (optional) additional kwargs that will be passed in the during the edit/update request
        :return: None
        :raises IllegalArgumentError: when the type or value of an argument provided is incorrect
        :raises APIError: in case an Error occurs
        """
        update_dict = {
            "id": self.id,
            "name": check_text(name, "name") or self.name,
            "description": check_text(description, "description") or "",
        }

        if kwargs:  # pragma: no cover
            update_dict.update(**kwargs)

        update_dict = clean_empty_values(update_dict=update_dict)
        try:
            response = self._client._request(
                "PUT",
                self._client._build_url("form", form_id=self.id),
                params=API_EXTRA_PARAMS["form"],
                json=update_dict,
            )
        except ForbiddenError:
            raise ForbiddenError("A form model with instances created cannot be edited")

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError(f"Could not update Form {self}", response=response)

        self.refresh(json=response.json().get("results")[0])

    def delete(self) -> None:
        """Delete this Form.

        :raises APIError: if delete was not successful.
        """
        response = self._client._request(
            "DELETE", self._client._build_url(self.url_detail_name, form_id=self.id)
        )

        if response.status_code != requests.codes.no_content:  # pragma: no cover
            raise APIError(f"Could not delete Form {self}", response=response)

    def instantiate(self, name: Optional[str], **kwargs) -> "Form":
        """Create a new Form instance based on a model."""
        if self.category != FormCategory.MODEL:
            raise APIError("Form should be of category MODEL")

        data_dict = {
            "name": check_text(name, "name") or self.name,
            "description": check_text(kwargs.get("description"), "description")
            or self.description,
            "contexts": check_list_of_base(kwargs.get("contexts"), Context, "contexts")
            or [],
        }

        url = self._client._build_url("form_instantiate", form_id=self.id)
        query_params = API_EXTRA_PARAMS["forms"]

        response = self._client._request(
            "POST", url, params=query_params, json=data_dict
        )

        if response.status_code != requests.codes.created:
            raise APIError(
                f"Could not instantiate this Form: {self}", response=response
            )

        instantiated_form = Form(response.json()["results"][0], client=self._client)
        return instantiated_form

    def clone(
        self, name: Optional[str], target_scope: Optional[Scope] = None, **kwargs
    ) -> Optional["Form"]:
        """Clone a new Form model based on a model.

        If target_scope is specified and different  from the scope of the form, then use the
        `clone_cross_scope` endpoint. Otherwise, use the basic `clone` endpoint.

        """
        if self.category != FormCategory.MODEL:
            raise APIError("Form should be of category MODEL")
        data_dict = {
            "name": check_text(name, "name") or f"{self.name}",
            "contexts": check_list_of_base(kwargs.get("contexts"), Context, "contexts")
            or [],
        }
        if "description" in kwargs:
            data_dict.update(
                {"description": check_text(kwargs.get("description"), "description")}
            )

        if not target_scope or target_scope.id == self.scope_id:
            response = self._client._request(
                "POST",
                self._client._build_url("form_clone", form_id=self.id),
                params=API_EXTRA_PARAMS["forms"],
                json=data_dict,
            )
        else:
            data_dict.update({"target_scope": check_base(target_scope, Scope, "scope")})
            response = self._client._request(
                "POST",
                self._client._build_url("form_clone_cross_scope", form_id=self.id),
                params=API_EXTRA_PARAMS["forms"],
                json=data_dict,
            )

        if response.status_code != requests.codes.created:
            raise APIError(f"Could not clone this Form: {self}", response=response)

        return Form(response.json()["results"][0], client=self._client)

    def activate(self):
        """Put the form to active.

        Model Forms can be put to inactive. When the Form model is not active the form cannot be
        instantiated. This methods sets the form to active.
        """
        if not self.active:
            url = self._client._build_url("form_activate", form_id=self.id)
            response = self._client._request("PUT", url=url)
            if response.status_code != requests.codes.ok:  # pragma: no cover
                raise APIError("Could not activate the form", response=response)

            # we need to do a full refresh here from the server as the
            # API of form/<id>/activate does not return the full object as response.
            self.refresh(
                url=self._client._build_url("form", form_id=self.id),
                extra_params=API_EXTRA_PARAMS.get(self.url_list_name),
            )

    def deactivate(self):
        """Put the form to inactive.

        Model Forms can be put to inactive. When the Form model is not active the form cannot be
        instantiated.
        """
        if self.active:
            url = self._client._build_url("form_deactivate", form_id=self.id)
            response = self._client._request("PUT", url=url)
            if response.status_code != requests.codes.ok:  # pragma: no cover
                raise APIError("Could not activate the form", response=response)

            # we need to do a full refresh here from the server as the
            # API of form/<id>/deactivate does not return the full object as response.
            self.refresh(
                url=self._client._build_url("form", form_id=self.id),
                extra_params=API_EXTRA_PARAMS.get(self.url_list_name),
            )

    def link_contexts(self, contexts: List[Union[Context, ObjectID]]):
        """
        Link a list of Contexts to a Form.

        :param contexts: a list of Context Objects or context_ids to link to the form.
        :raises APIError: in case an Error occurs when linking
        """
        data = {"contexts": check_list_of_base(contexts, Context)}
        url = self._client._build_url("form_link_contexts", form_id=self.id)
        query_params = API_EXTRA_PARAMS.get(self.url_list_name)
        response = self._client._request(
            "POST", url=url, params=query_params, json=data
        )
        if response.status_code != requests.codes.ok:
            raise APIError(
                "Could not link the specific contexts to the form",
                response=response,
            )
        self.refresh(json=response.json()["results"][0])

    def unlink_contexts(self, contexts: List[Union[Context, ObjectID]]):
        """
        Unlink a list of Contexts from a Form.

        :param contexts: a list of Context Objects or context_ids to unlink from the form.
        :raises APIError: in case an Error occurs when unlinking
        """
        data = {"contexts": check_list_of_base(contexts, Context)}
        url = self._client._build_url("form_unlink_contexts", form_id=self.id)
        query_params = API_EXTRA_PARAMS.get(self.url_list_name)
        response = self._client._request(
            "POST", url=url, params=query_params, json=data
        )
        if response.status_code != requests.codes.ok:
            raise APIError(
                "Could not unlink the specific contexts from the form",
                response=response,
            )
        self.refresh(json=response.json()["results"][0])

    def set_status_assignees(self, statuses: List[dict]):
        """
        Set a list of assignees on each status of a Form.

        :param statuses: a list of dicts, each one contains the status_id and the list of
        assignees. Available fields per dict:
            :param status: Status object
            :param assignees: List of User objects
        :raises APIError: in case an Error occurs when setting the status assignees

        Example
        -------
        When the `Form` is known, one can easily access its status forms ids and build a
        dictionary, as such:

        >>> for status_form in form.status_forms:
        >>>    status_dict = {
        >>>        "status": status_form.status,
        >>>        "assignees": [user_1, user_2]
        >>>    }
        >>>     status_assignees_list.append(status_dict)
        >>> self.form.set_status_assignees(statuses=status_assignees_list)

        """
        from pykechain.models import User

        check_list_of_dicts(
            statuses,
            "statuses",
            [
                "status",
                "assignees",
            ],
        )
        for status in statuses:
            status["status"] = check_base(status["status"], Status)
            status["assignees"] = check_list_of_base(status["assignees"], User)
        url = self._client._build_url("form_set_status_assignees", form_id=self.id)
        query_params = API_EXTRA_PARAMS.get(self.url_list_name)
        response = self._client._request(
            "POST", url=url, params=query_params, json=statuses
        )
        if response.status_code != requests.codes.ok:
            raise APIError(
                "Could not set the list of status assignees to the form",
                response=response,
            )
        self.refresh(json=response.json()["results"][0])

    def possible_transitions(self) -> List[Transition]:
        """Retrieve the possible transitions that may be applied on the Form.

        It will return the Transitions from the associated workflow are can be applied
        on the Form in the current status.
        :returns: A list with possible Transitions that may be applied on the Form.
        """
        workflow = self._client.workflow(id=self._workflow["id"])
        return workflow.transitions

    def apply_transition(self, transition: Transition):
        """Apply the transition to put the form in another state following a transition.

        Apply transition is to put the Form in another state. Only transitions that
        can apply to the form should have the 'from_status' to the current state. (or
        it is a Global transition). If applied the Form will be set in the 'to_state' of
        the Transition.

        :param transition: a Transition object belonging to the workflow of the Form
        :raises APIError: in case an Error occurs when applying the transition

        Example
        -------
        When the `Form` and `Workflow` is known, one can easily apply a transition on it, as such:

        >>> transition = workflow.transition("In progress")
        >>> form.apply_transition(transition=transition)

        """
        check_base(transition, Transition)

        url = self._client._build_url(
            "form_apply_transition", form_id=self.id, transition_id=transition.id
        )
        query_params = API_EXTRA_PARAMS.get(self.url_list_name)
        response = self._client._request(
            "POST",
            url=url,
            params=query_params,
        )
        if response.status_code != requests.codes.ok:
            raise APIError(
                "Could not transition the form",
                response=response,
            )
        self.refresh(json=response.json()["results"][0])

    def has_part(self, part: Part) -> bool:
        """Return boolean if given Part is part of the Form tree.

        Based on the Parts category, either the model or the instance tree is checked.

        :param part: a Part object
        :raises IllegalArgumentError: in case not a `Part` object is used when calling the method
        :raises APIError: in case an Error occurs when checking whether `Form` contains the `Part`
        """
        part_id = check_base(part, Part)

        url = self._client._build_url(
            "forms_has_part", form_id=self.id, part_id=part_id
        )
        response = self._client._request("GET", url=url)
        if response.status_code != requests.codes.ok:
            raise APIError(
                f"Could not process whether `Form` {self.id} has part {part_id}",
                response=response,
            )
        return response.json()["results"][0]["has_part"]

    def set_prefill_parts(self, prefill_parts: dict) -> None:
        """Set the prefill_parts on the Form.

        Prefill parts are insatnces of a part model that are stored on the
        Form Model (only for models) and these instances of the parts are
        instantiated by the backend when the Form is also instantiated.

        Set the prefill_parts on the form collection model using a special
        structure. The structure is jsonschema validated by the backend before
        the `Form` is updated.

        The prefill_parts on the form should be in the following structure:
        ```
        {
            <uuid of the part_model> : [
                {
                    # an individual instance with 'partname'='name'
                    name: null_string,
                    part_properties: [
                        {
                            property_id: <uuid>,
                            value: <null_string>
                        },{...}
                    ]
                }, {...}
            ],
            <uuid of another part_model>: [...]
        }
        ```
        :param prefill_parts: dictionary of `<part_model_id>: [array of instances to create]`
        :raises APIError: When the prefill_parts is tried to be set on Form Instances
        :raises APIError: WHen the update of the prefill_parts on the Form does not succeed.
        """
        # only works for Form Models.
        if not self.is_model:
            raise APIError(
                "The update of the prefill parts is only applicable to Form Models."
            )

        payload = dict(
            prefill_parts=check_json(
                value=prefill_parts,
                schema=form_collection_prefill_parts_schema,
                key="prefill_parts",
            )
        )

        url = self._client._build_url("form", form_id=self.id)
        query_params = API_EXTRA_PARAMS.get(self.url_list_name)

        response = self._client._request(
            "PATCH", url=url, params=query_params, json=payload
        )
        if response.status_code != requests.codes.ok:
            raise APIError(
                "Could not update the prefill_parts dictionary on the form collection",
                response=response,
            )
        self.refresh(json=response.json()["results"][0])

    def workflows_compatible_with_scope(self, scope: Scope):
        """Return workflows from target scope that are compatible with source workflow.

        :param scope: a Scope object
        :raises IllegalArgumentError: in case not a `Scope` object is used when calling the method
        :raises APIError: in case an Error occurs

        """
        scope_id = check_base(scope, Scope)

        url = self._client._build_url(
            "forms_compatible_within_scope", form_id=self.id, scope_id=scope_id
        )
        response = self._client._request("GET", url=url)
        if response.status_code != requests.codes.ok:
            raise APIError(
                "Could not retrieve the compatible workflows",
                response=response,
            )
        return response.json()["results"]



from pykechain.enums import (
    Alignment,
    KEChainPageLabels,
    KEChainPages,
    ScopeStatus,
    SidebarAccessLevelOptions,
    SidebarItemAlignment,
    SubprocessDisplayMode,
    URITarget,
)
from pykechain.exceptions import IllegalArgumentError, NotFoundError
from pykechain.models import Scope
from pykechain.models.sidebar.sidebar_button import SideBarButton
from pykechain.models.sidebar.sidebar_card import SideBarCard
from pykechain.models.sidebar.sidebar_manager import SideBarManager
from tests.classes import TestBetamax


class TestSideBar(TestBetamax):
    def setUp(self):
        super().setUp()

        self.scope = self.project.clone(asynchronous=False)  # type: Scope
        del self.project

        self.manager = SideBarManager(scope=self.scope)
        self.scope_uri = f"#/scopes/{self.scope.id}"
        self.scope_activity_uri = f"{self.scope_uri}/{SubprocessDisplayMode.ACTIVITIES}"

        self.default_button_config = {
            "title": "new button",
            "icon": "bookmark",
            "uri": f"{self.scope_uri}/{KEChainPages.DATA_MODEL}",
        }

    def tearDown(self):
        self.scope.delete()
        super().tearDown()

    def test_manager(self):
        side_bar_manager = SideBarManager(scope=self.scope)

        self.assertIsInstance(side_bar_manager, SideBarManager)

    def test_singleton_manager_per_scope(self):
        manager_1 = SideBarManager(scope=self.scope)
        manager_2 = self.scope.side_bar()

        self.assertTrue(
            manager_1 is manager_2,
            msg="There are 2 different manager objects while the same object is expected.",
        )

    def test_loading_of_existing_buttons(self):
        for scope in self.client.scopes(status=ScopeStatus.ACTIVE):
            scope.side_bar()

    def test_override_sidebar_property(self):
        value = self.manager.override_sidebar

        self.assertIsInstance(value, bool)
        self.assertFalse(value)

        self.manager.override_sidebar = True

        self.assertTrue(self.manager.override_sidebar)

    def test_create_button(self):
        new_button = self.manager.create_button(**self.default_button_config)

        self.assertIsInstance(new_button, SideBarButton)
        self.assertTrue(new_button is self.manager[-1])
        self.assertIsInstance(new_button.__repr__(), str)

    # noinspection PyTypeChecker
    def test_create_button_incorrect_arguments(self):
        with self.assertRaises(IllegalArgumentError):
            SideBarButton(self.manager, order="3")

        with self.assertRaises(IllegalArgumentError):
            SideBarButton(self.manager, order=1, title=False)

        with self.assertRaises(IllegalArgumentError):
            SideBarButton(self.manager, order=1, title="Button", icon=32)

        with self.assertRaises(IllegalArgumentError):
            SideBarButton(self.manager, order=1, title="Button", icon="pennant", uri=44)

        with self.assertRaises(IllegalArgumentError):
            SideBarButton(
                self.manager,
                order=1,
                title="Button",
                icon="pennant",
                uri="http://www.google.com",
                uri_target="any",
            )

        with self.assertRaises(IllegalArgumentError):
            SideBarButton(
                self.manager,
                order=1,
                title="Button",
                icon="pennant",
                uri="http://www.google.com",
                icon_mode="best",
            )

        with self.assertRaises(IllegalArgumentError):
            SideBarButton(
                self.manager,
                order=1,
                title="Button",
                icon="pennant",
                uri="http://www.google.com",
                random="unsupported keyword",
            )

    def test_edit_button(self):
        custom_name = "Custom Dutch name"
        new_button = self.manager.create_button(**self.default_button_config)

        new_button.edit(
            display_icon="pennant",
            displayName_nl=custom_name,
        )
        new_button.refresh()

        self.assertEqual("pennant", new_button.display_icon)
        self.assertEqual(
            custom_name, new_button._other_attributes.get("displayName_nl")
        )

    def test_delete_button(self):
        new_button = self.manager.create_button(**self.default_button_config)
        new_button.delete()

        self.assertNotIn(new_button, self.manager)

    def test_get_button(self):
        new_button = self.manager.create_button(**self.default_button_config)

        button_from_index = self.manager[0]
        button_from_title = self.manager["new button"]
        button_from_button = self.manager[new_button]

        self.assertIs(new_button, button_from_index)
        self.assertIs(new_button, button_from_title)
        self.assertIs(new_button, button_from_button)

        with self.assertRaises(IndexError):
            # noinspection PyStatementEffect
            self.manager[3]

        with self.assertRaises(NotFoundError):
            # noinspection PyStatementEffect
            self.manager["unused button title"]

    def test_remove_button(self):
        new_button = self.manager.create_button(**self.default_button_config)

        self.manager.remove(new_button)

        self.assertTrue(len(self.manager) == 0)

    def test_insert_button(self):
        button_1 = self.manager.create_button(order=0, **self.default_button_config)
        button_2 = self.manager.create_button(order=1, **self.default_button_config)
        button_3 = self.manager.create_button(order=2, **self.default_button_config)

        self.manager.insert(0, button_2)

        self.assertListEqual(
            [button_2, button_1, button_3],
            list(self.manager),
        )

    def test_bulk_add_buttons(self):
        buttons = [
            {
                "displayName": "activity button",
                "displayIcon": "bookmark",
                "uri": f"{self.scope_activity_uri}/{self.scope.activities()[0].id}",
            },
            {
                "displayName": "ke-chain page",
                "displayIcon": "university",
                "uri": f"{self.scope_uri}/{KEChainPages.EXPLORER}",
            },
            {
                "displayName": "external button",
                "displayIcon": "google",
                "uri": "https://www.google.com",
            },
        ]
        new_buttons = self.manager.add_buttons(
            buttons=buttons,
            override_sidebar=True,
        )

        self.assertIsInstance(new_buttons, list)
        self.assertTrue(all(isinstance(b, SideBarButton) for b in new_buttons))
        self.assertTrue(len(self.manager) == 3)

    def test_context_manager(self):
        with SideBarManager(scope=self.scope) as bulk_creation_manager:
            bulk_creation_manager.override_sidebar = True

            bulk_creation_manager.add_task_button(
                activity=self.scope.activities()[0],
                icon="bookmark",
            )
            bulk_creation_manager.add_external_button(
                url="https://www.google.com",
                icon="google",
                title="Google",
            )
            bulk_creation_manager.add_ke_chain_page(
                page_name=KEChainPages.WORK_BREAKDOWN,
                icon="sitemap",
            )

            self.assertEqual(
                0,
                len(self.scope.options.get("customNavigation", list())),
                msg="During bulk creation of buttons, KE-chain should not be updated yet.",
            )

        self.assertEqual(
            3,
            len(self.scope.options["customNavigation"]),
            msg="After closing the context `with`, update should be performed.",
        )

        updated_side_bar_buttons = self.scope.options.get("customNavigation")
        self.assertTrue(
            len(updated_side_bar_buttons) == 3,
            msg="At the end of bulk creation, the buttons must have been created.",
        )

        self.assertTrue(self.scope.options.get("overrideSideBar"))

    def test_load_buttons(self):
        # setup: create a custom button
        with self.scope.side_bar() as sbm:
            sbm.override_sidebar = True
            sbm.add_ke_chain_page(KEChainPages.WORK_BREAKDOWN)

        # Reload side-bar from KE-chain
        self.scope.side_bar().refresh()

        # testing: delete original button (this tests the "load") and add another one
        with self.scope.side_bar() as sbm:
            for button in sbm:
                sbm.delete_button(button)
            sbm.override_sidebar = True
            sbm.add_ke_chain_page(KEChainPages.TASKS)

        self.assertEqual(1, len(self.scope.options.get("customNavigation")))

    def test_attributes_button(self):
        display_name_nl = "Het data model"
        button_title = "New button"

        # Check whether all attributes are retrieved and set
        with self.scope.side_bar() as sbm:
            sbm.add_ke_chain_page(
                page_name=KEChainPages.DATA_MODEL,
                title=button_title,
                displayName_nl=display_name_nl,
            )

        # Reload side-bar from KE-chain
        self.scope.side_bar().refresh()

        sbm = self.scope.side_bar()
        found = False
        for button in sbm:  # type: SideBarButton
            if button._other_attributes.get("displayName_nl") == display_name_nl:
                found = True

        self.assertTrue(found)

    def test_attributes_sidebar(self):
        sbm = self.scope.side_bar()

        # Check starting condition
        self.assertFalse(sbm.override_sidebar)

        # Set value
        sbm.override_sidebar = True

        # Reload side-bar from KE-chain
        sbm.refresh()

        # Test attributes
        self.assertTrue(sbm.override_sidebar)

    def test_alignment(self):
        # Check whether all attributes are retrieved and set
        with self.scope.side_bar() as sbm:
            sbm.add_ke_chain_page(
                page_name=KEChainPages.CONTEXTS,
                alignment=SidebarItemAlignment.TOP,
            )
            sbm.add_ke_chain_page(
                page_name=KEChainPages.FORMS,
                alignment=SidebarItemAlignment.BOTTOM,
            )

        # Reload side-bar from KE-chain
        self.scope.side_bar().refresh()

        found = [False, False]

        sbm = self.scope.side_bar()
        for button in sbm:  # type: SideBarButton
            if button.display_name == KEChainPageLabels[KEChainPages.FORMS]:
                self.assertEqual(button.alignment, SidebarItemAlignment.BOTTOM)
                found[0] = True
            if button.display_name == KEChainPageLabels[KEChainPages.CONTEXTS]:
                self.assertEqual(button.alignment, SidebarItemAlignment.TOP)
                found[1] = True

        self.assertTrue(all(found))

    def test_side_bar_permissions(self):
        # Check whether all attributes are retrieved and set
        with self.scope.side_bar() as sbm:
            sbm.add_ke_chain_page(
                page_name=KEChainPages.FORMS,
                minimum_access_level=SidebarAccessLevelOptions.IS_MEMBER,
            )
            sbm.add_ke_chain_page(
                page_name=KEChainPages.CONTEXTS,
                minimum_access_level=SidebarAccessLevelOptions.IS_LEAD_MEMBER,
            )
            sbm.add_ke_chain_page(
                page_name=KEChainPages.CATALOG_FORMS,
                minimum_access_level=SidebarAccessLevelOptions.IS_SUPERVISOR,
            )
            sbm.add_ke_chain_page(
                page_name=KEChainPages.WORKFLOWS,
                minimum_access_level=SidebarAccessLevelOptions.IS_MANAGER,
            )
            sbm.add_ke_chain_page(
                page_name=KEChainPages.DETAIL,
            )

        # Reload side-bar from KE-chain
        self.scope.side_bar().refresh()

        found = [False, False, False, False, False]

        sbm = self.scope.side_bar()
        for button in sbm:  # type: SideBarButton
            if button.display_name == KEChainPageLabels[KEChainPages.FORMS]:
                self.assertTrue(
                    button.minimum_access_level == SidebarAccessLevelOptions.IS_MEMBER
                )
                found[0] = True
            if button.display_name == KEChainPageLabels[KEChainPages.CONTEXTS]:
                self.assertTrue(
                    button.minimum_access_level
                    == SidebarAccessLevelOptions.IS_LEAD_MEMBER
                )
                found[1] = True
            if button.display_name == KEChainPageLabels[KEChainPages.CATALOG_FORMS]:
                self.assertTrue(
                    button.minimum_access_level
                    == SidebarAccessLevelOptions.IS_SUPERVISOR
                )
                found[2] = True
            if button.display_name == KEChainPageLabels[KEChainPages.WORKFLOWS]:
                self.assertTrue(
                    button.minimum_access_level == SidebarAccessLevelOptions.IS_MANAGER
                )
                found[3] = True
            if button.display_name == KEChainPageLabels[KEChainPages.DETAIL]:
                self.assertTrue(
                    button.minimum_access_level == SidebarAccessLevelOptions.IS_MEMBER
                )
                found[4] = True

        self.assertTrue(all(found))

    def test_sidebar_card_object(self):
        """Aim to create a sidebard card."""
        target_dict = {
            "itemType": "CARD",
            "align": "top",
            "showCloseAction": True,
            "showActionButton": True,
            "actionButtonUri": "https://changelog.md",
            "actionButtonUriTarget": "_new",
            "actionButtonName": "Discover more",
            "actionButtonName_nl": "Ontdek meer",
            "displayText": "This project ...",
            "displayText_nl": "Dit project ...",
            "displayTextAlign": "left",
            "showBackground": True,
            "minimumAccessLevel": "is_member",
            "maximumAccessLevel": "is_supervisor",
        }

        sidebar_card = SideBarCard(
            side_bar_manager=self.manager,
            alignment=SidebarItemAlignment.TOP,
            show_close_action=True,
            show_action_button=True,
            action_button_name="Discover more",
            action_button_uri="https://changelog.md",
            action_button_uri_target=URITarget.NEW,
            display_text="This project ...",
            display_text_align=Alignment.LEFT,
            show_background=True,
            minimum_access_level=SidebarAccessLevelOptions.IS_MEMBER,
            maximum_access_level=SidebarAccessLevelOptions.IS_SUPERVISOR,
            displayText_nl="Dit project ...",
            actionButtonName_nl="Ontdek meer",
        )

        card_dict = sidebar_card.as_dict()

        self.maxDiff = None
        self.assertDictEqual(target_dict, card_dict)

    def test_sidebar_card_creation_inside_the_manager(self):
        with self.scope.side_bar() as sbm:
            sbm.create_card(
                display_text="Foo Bar",
                show_action_button=False,
            )

        # Reload side-bar from KE-chain
        self.scope.side_bar().refresh()

        sbm = self.scope.side_bar()
        self.assertTrue(len(sbm) == 1)
        for item in sbm:
            self.assertEqual(item.display_text, "Foo Bar")
            self.assertFalse(hasattr(item, "order"))
            self.assertFalse(item.show_action_button)



from typing import Dict, Optional

from pykechain.enums import SidebarType


class SideBarItem:
    """
    A base class for all various sidebar 'widgets'.

    :cvar allowed_attributes: allowed additional attributed provided as options alongside
        the specifically allowed ones.
    :cvar item_type: the item type of this class. Defaults to a BUTTON.
    """

    _allowed_attributes = []
    _item_type = SidebarType.BUTTON

    def __repr__(self) -> str:
        if hasattr(self, "title"):
            name = getattr(self, "title")
        elif hasattr(self, "display_text"):
            name = getattr(self, "display_text")
        else:
            name = "<no name>"
        return f"<pyke {self.__class__.__name__}: '{name}'>"

    def refresh(self, json: Optional[Dict] = None) -> None:
        """
        Refresh the object in-place using the provided json.

        :param json: the json response to construct the :class:`SideBarButton` from
        :type json: dict
        :return: None
        """
        self.__init__(side_bar_manager=self._manager, json=json or self.as_dict())

    def edit(self, **kwargs) -> None:
        """
        Edit the details of the button.

        :param kwargs:
        :return: None
        """
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            elif key in self._allowed_attributes:
                self._other_attributes[key] = value

        self._manager._update()

    def delete(self) -> None:
        """
        Delete the side-bar button from the side-bar.

        :return: None
        """
        self._manager.delete_button(key=self)

    def as_dict(self) -> Dict:
        """Represent the sidebar object as a dictionary."""
        raise NotImplementedError("The method `as_dict()` need to be implemented.")
